theory Safety_Checker
imports 
  "$AFP/Check_Monad/Check_Monad"
  "~~/src/HOL/Library/RBT_Mapping"
  Formula
  LTS
begin

type_synonym art_node = string

text {*
  Below, @{typ 'l} and @{typ 'tr} will denote the types of locations and transitions.
*}

datatype 'tr art_edge = is_cover_node: Cover art_node | Children "('tr \<times> art_node) list"

record ('f,'v,'t,'l,'tr) art =
  initial_node :: art_node
  nodes :: "art_node set"
  edge :: "art_node \<Rightarrow> 'tr art_edge"
  node_location :: "art_node \<Rightarrow> 'l"
  node_invariant :: "art_node \<Rightarrow> ('f,'v,'t) formula"

definition transition_formula_of ::
  "('f,'v,'t,'l,'tr) lts \<Rightarrow> 'tr \<Rightarrow> ('f,'v,'t) transition_formula" where
  "transition_formula_of P \<tau> = trans_formula (transition_of P \<tau>)"

context pretheory begin

definition transition_entailment ::
  "('f,'v,'t,'l,'tr) lts \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> 'tr \<Rightarrow> ('f,'v,'t) formula \<Rightarrow> bool" where
  "transition_entailment P \<phi> \<tau> \<psi> \<equiv>
   rename_vars Pre \<phi> \<and>\<^sub>f transition_formula_of P \<tau> \<longrightarrow>\<^sub>f rename_vars Post \<psi>"

end

locale pre_art_body = "pretheory" T
  for T :: "('f,'t,'d) theory"
  and P :: "('f,'v,'t,'l,'tr) lts"
  and A :: "('f,'v,'t,'l,'tr) art"
  and err :: "'l"
begin

definition "art \<equiv> \<forall>a \<in> nodes A. formula (node_invariant A a)"

definition "art_checker \<equiv> check art (shows ''ill-formed invariant'')"

lemma art_checker[intro]: "isOK art_checker \<Longrightarrow> art" unfolding art_checker_def by auto

definition "initial_cond \<equiv>
  initial_node A \<in> nodes A \<and>
  node_invariant A (initial_node A) = form_True"

definition "simulation_cond \<equiv>
  node_location A (initial_node A) = initial P \<and>
  (\<forall> a children l \<tau> \<phi> l'.
    a \<in> nodes A
    \<longrightarrow> edge A a = Children children
    \<longrightarrow> l = node_location A a
    \<longrightarrow> \<tau> \<in> transitions P
    \<longrightarrow> transition_of P \<tau> = Transition l \<phi> l'
    \<longrightarrow> (\<exists> b \<in> nodes A. (\<tau>, b) \<in> set children \<and> node_location A b = l'))"

definition "cover_edges_cond \<equiv> \<forall> a b.
    a \<in> nodes A
    \<longrightarrow> edge A a = Cover b
    \<longrightarrow> b \<in> nodes A \<and>
      node_location A a = node_location A b \<and>
      \<not> is_cover_node (edge A b) \<and>
      node_invariant A a \<longrightarrow>\<^sub>f node_invariant A b"

definition "children_edges_cond \<equiv> \<forall> a children \<tau> b.
    a \<in> nodes A
    \<longrightarrow> edge A a = Children children
    \<longrightarrow> (\<tau>,b) \<in> set children
    \<longrightarrow> transition_entailment P (node_invariant A a) \<tau> (node_invariant A b)"

definition "error_cond \<equiv> \<forall> a.
    a \<in> nodes A \<longrightarrow> node_location A a = err \<longrightarrow> node_invariant A a \<longrightarrow>\<^sub>f form_False"

definition "well_formed \<equiv>
  children_edges_cond \<and> cover_edges_cond \<and> initial_cond \<and> error_cond \<and> simulation_cond \<and> art"

fun corresponds_to :: "('v,'t,'d,'l) state \<Rightarrow> art_node \<Rightarrow> bool" where
  "corresponds_to (State \<alpha> l) a  = (node_location A a = l \<and>
    satisfies \<alpha> (node_invariant A a))"

lemma pre_valuation:
  "satisfies (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (rename_vars Pre \<phi>) = satisfies \<alpha>1 \<phi>"
  by auto

lemma post_valuation:
  "satisfies (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) (rename_vars Post \<phi>) = satisfies \<alpha>2 \<phi>"
  by auto

lemma well_formed_sound:
  assumes well_formed
  and safety: "safety_prog (Safety_Prog err P)"
  shows "safe (Safety_Prog err P)"
  unfolding safe_def safety_prog.case
proof (rule ccontr)
  assume "\<not> (\<forall>\<alpha>. assignment \<alpha> \<longrightarrow> State \<alpha> err \<notin> reachable_states P)"
  then obtain \<alpha> where \<alpha>: "assignment \<alpha>" and reach: "State \<alpha> err \<in> reachable_states P" by auto
  then obtain init
    where init: "init \<in> initial_states P"
      and steps: "(init, State \<alpha> err) \<in> (transition P)^*" by (auto simp: reachable_states_def)
  from assms[unfolded well_formed_def]
  have ic: initial_cond and ec: error_cond
   and sc: simulation_cond and cec: children_edges_cond
   and cc: cover_edges_cond and art: art by auto
  from art have [intro]: "\<And> a. a \<in> nodes A \<Longrightarrow> formula (node_invariant A a)"
    unfolding art_def by auto
  obtain l \<alpha>0 where "init = State \<alpha>0 l" by (cases init, auto)
  with init[unfolded initial_states_def] have "init = State \<alpha>0 (initial P)" by auto
  with ic sc have ic: "initial_node A \<in> nodes A" "corresponds_to init (initial_node A)"
    by (auto simp: simulation_cond_def initial_cond_def)
  {
    fix s1 s2 a1
    assume t: "(s1,s2) \<in> transition P" and s1a1: "corresponds_to s1 a1" and a1: "a1 \<in> nodes A"
    obtain \<alpha>1 l1 where s1: "s1 = State \<alpha>1 l1" by (cases s1, auto)
    obtain \<alpha>2 l2 where s2: "s2 = State \<alpha>2 l2" by (cases s2, auto)
    from t[unfolded transition_def s1 s2] obtain \<tau> \<phi> \<beta>
    where trans: "\<tau> \<in> transitions P" "transition_of P \<tau> = Transition l1 \<phi> l2"
      and transf: "satisfies (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>) \<phi>"
      and \<alpha>1: "assignment \<alpha>1"
      and \<alpha>2: "assignment \<alpha>2"
      and \<beta>: "assignment \<beta>"
      by blast
    from safety[simplified] trans have \<phi>: "formula \<phi>"
      unfolding lts_def wt_transition_def by auto
    def \<alpha> \<equiv> "(pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>)"
    from \<alpha>1 \<alpha>2 \<beta> have ass: "assignment (pre_post_inter_valuation \<alpha>1 \<alpha>2 \<beta>)"
      by (auto simp: assignment_def, case_tac "x", auto)
    hence \<alpha>: "assignment \<alpha>" unfolding \<alpha>_def .
    have "\<exists> a1 tas. corresponds_to s1 a1 \<and> edge A a1 = Children tas \<and> a1 \<in> nodes A"
    proof (cases "edge A a1")
      case (Children tas)
      with s1a1 a1 show ?thesis by blast
    next
      case (Cover b)
      from cc[unfolded cover_edges_cond_def, rule_format, OF a1 Cover] obtain tas where
        b: "b \<in> nodes A" and loc: "node_location A a1 = node_location A b"
        and edge: "edge A b = Children tas"
        and impl: "node_invariant A a1 \<longrightarrow>\<^sub>f node_invariant A b" by (cases "edge A b", auto)
      show ?thesis
      proof (rule exI[of _ b], rule exI[of _ tas], intro conjI)
        from s1a1[unfolded s1, simplified]
        have loc: "node_location A b = l1" and alpha: "satisfies \<alpha>1 (node_invariant A a1)" 
          using loc by auto
        from impliesD[OF \<alpha>1 _ _ impl alpha] a1 b
        have "satisfies \<alpha>1 (node_invariant A b)" by auto
        with loc show "corresponds_to s1 b" unfolding s1 by auto
      qed (insert b edge, auto)
    qed
    then obtain a1 tas where s1a1: "corresponds_to s1 a1"
      and edge: "edge A a1 = Children tas"
      and a1: "a1 \<in> nodes A" by blast
    from s1a1[unfolded s1] have l1: "l1 = node_location A a1"
      and \<alpha>1: "satisfies \<alpha>1 (node_invariant A a1)" by auto
    from sc[unfolded simulation_cond_def, THEN conjunct2,
      rule_format, OF a1 edge l1 trans] obtain b where
      b: "b \<in> nodes A" and mem: "(\<tau>, b) \<in> set tas" and l2: "node_location A b = l2" by auto
    from cec[unfolded children_edges_cond_def, rule_format, OF a1 edge mem]
    have "transition_entailment P (node_invariant A a1) \<tau> (node_invariant A b)" .
    from this[unfolded transition_entailment_def transition_formula_of_def] trans(2)
    have "rename_vars Pre (node_invariant A a1) \<and>\<^sub>f \<phi> \<longrightarrow>\<^sub>f rename_vars Post (node_invariant A b)" by simp
    from impliesD[OF _ _ _ this satisfies_andI[OF _ transf], unfolded formula_rename_vars formula_and pre_valuation post_valuation, 
      OF ass conjI[OF _ \<phi>] _ \<alpha>1]
    have "satisfies \<alpha>2 (node_invariant A b)" using a1 b by auto
    with l2 have "corresponds_to s2 b" unfolding s2 by simp
    with b have "\<exists> a2 \<in> nodes A. corresponds_to s2 a2" by blast
  } note one_step_simulation = this
  {
    fix s1 s2 a1
    assume t: "(s1,s2) \<in> (transition P)^*" and mem: "a1 \<in> nodes A" and s1a1: "corresponds_to s1 a1"
    from t have "\<exists> a2 \<in> nodes A. corresponds_to s2 a2" 
    proof (induct rule: rtrancl_induct)
      case (step s2 s3)
      from step(3) obtain a2 where a2: "a2 \<in> nodes A" and s2a2: "corresponds_to s2 a2" by blast
      from one_step_simulation[OF step(2) s2a2 a2] show ?case .
    qed (insert s1a1 mem, auto)
  } note many_step_simulation = this
  from many_step_simulation[OF steps ic]
  obtain a_err where a_err: "a_err \<in> nodes A" and err: "corresponds_to (State \<alpha> err) a_err" by auto
  then have err: "node_location A a_err = err" "satisfies \<alpha> (node_invariant A a_err)" by auto
  from ec[unfolded error_cond_def, rule_format, OF a_err err(1)]
  have "node_invariant A a_err \<longrightarrow>\<^sub>f form_False" .
  from impliesD[OF \<alpha> _ _ this err(2)] a_err show False by auto
qed

end

record ('h,'l,'tr) hinter =
  nodes :: "art_node list"
  trans_P_list :: "'l \<Rightarrow> ('tr \<times> 'l) list"
  error_hints :: "art_node \<Rightarrow> 'h list option"
  cover_hints :: "art_node \<Rightarrow> 'h list option"
  transition_hints :: "art_node \<Rightarrow> 'h list list option"

locale pre_art_checker = pretheory T
  for T :: "('f::show,'t::show,'d) theory" +
  fixes tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2 :: "('h,'f,'v trans_var,'t) theory_checker"

locale pre_art_checker_body =
  pre_art_checker T tc tc2 + pre_art_body T P A err
  for T :: "('f::show,'t::show,'d) theory"
  and tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2
  and P :: "('f,'v,'t,'l::show,'tr::show) lts"
  and A
  and err +
  fixes H :: "('h,'l,'tr) hinter"
begin

definition simulation_cond_checker :: "shows check" where
  "simulation_cond_checker \<equiv> do {
    check (node_location A (initial_node A) = initial P)
      (shows ''initial node of ART does not correspond to initial node of program'');
    check_allm (\<lambda> a.
      case edge A a
      of Cover _ \<Rightarrow> succeed
      | Children children \<Rightarrow>
          let l = node_location A a in
          check_allm
            (\<lambda> (\<tau>,l').
              check (\<exists> (\<tau>',b) \<in> set children. \<tau>' = \<tau> \<and> node_location A b = l' \<and> b \<in> art.nodes A)
            (shows ''could not find matching transition in ART for node '' o
             shows a o shows '' and transition '' o shows \<tau>))
            (trans_P_list H l)
    ) (nodes H)
  } <+? (\<lambda> s. shows ''could not ensure simulation condition of ART\<newline>'' o s)"

definition error_cond_checker :: "shows check" where
  "error_cond_checker \<equiv> do {
     check_allm (\<lambda> a.
       if node_location A a = err then
         case error_hints H a of None \<Rightarrow>
           error (shows ''could not find hint for checking error condition of node '' o shows a)
         | Some hints \<Rightarrow> pre_theory_checker.formula_checker tc hints (node_invariant A a) form_False
       else succeed)
       (nodes H)
   } <+? (\<lambda> s. shows ''could not ensure error condition of ART\<newline>'' o s)"

definition cover_edges_cond_checker :: "shows check" where
  "cover_edges_cond_checker \<equiv> do {
     check_allm (\<lambda> a.
       case edge A a of Children _ \<Rightarrow> succeed
       | Cover b \<Rightarrow> do {
           check (b \<in> art.nodes A) (shows ''target node is not listed as art-node'');
           check (node_location A a = node_location A b) (shows ''node-locations differ'');
           check (\<not> is_cover_node (edge A b)) (shows ''target node must not have cover edge'');
           case cover_hints H a of None \<Rightarrow> error (shows ''there is no cover hint present'')
           | Some hints \<Rightarrow> pre_theory_checker.formula_checker tc hints (node_invariant A a) (node_invariant A b)
       } <+? (\<lambda> s. shows ''problem in checking cover edge of node '' o shows a o shows_nl o s)
     ) (nodes H)
   } <+? (\<lambda> s. shows ''could not ensure cover edge condition of ART\<newline>'' o s)" 

definition initial_cond_checker :: "shows check" where
  "initial_cond_checker \<equiv> do {
    let init = initial_node A;
    check (init \<in> art.nodes A)
      (shows ''initial node of A ('' o shows init o shows '') is not mentioned as node of ART''
      );
    check (node_invariant A init = form_True)
     (shows ''the node invariant for the initial ART node ('' o shows init o shows '') must be TRUE'')
  }"

definition children_edges_cond_checker :: "shows check" where
  "children_edges_cond_checker \<equiv> do {
     check_allm (\<lambda> a.
       case edge A a of Cover _ \<Rightarrow> succeed
       | Children children \<Rightarrow> do {
           case transition_hints H a of None \<Rightarrow> error (shows ''there is no transition hint present'')
           | Some hintss \<Rightarrow> do {
               check (length hintss = length children)
                 (shows ''the number of hints differs from the number of children'');
               check_allm (\<lambda> ((\<tau>,b),hints).
                 pre_theory_checker.formula_checker tc2 hints
                   (rename_vars_formula Pre (node_invariant A a) \<and>\<^sub>f transition_formula_of P \<tau>)
                   (rename_vars_formula Post (node_invariant A b)) <+?
                 (\<lambda> s. shows ''problem in checking transition '' o shows a o shows '' --'' o
                   shows \<tau> o shows ''--> '' o shows b o shows_nl o s)
               ) (zip children hintss)
             }
       } <+? (\<lambda> s. shows ''problem in checking transitions of node '' o shows a o shows_nl o s)
     ) (nodes H)
   } <+? (\<lambda> s. shows ''could not ensure transition edge condition of ART\<newline>'' o s)" 

definition art_safety_checker_main  :: "shows check" where
  "art_safety_checker_main \<equiv> do {
      art_checker;
      initial_cond_checker;
      simulation_cond_checker;
      error_cond_checker;
      cover_edges_cond_checker;
      children_edges_cond_checker
   } <+? (\<lambda> s. shows ''could not ensure safety via ART:\<newline>'' o s)"

end

locale art_checker =
  pre_art_checker T tc tc2 +
  tc: theory_checker T tc +
  tc2: theory_checker T tc2
  for T :: "('f::show,'t::show,'d) theory"
  and tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2

locale art_checker_body =
  art_checker T tc tc2 +
  pre_art_checker_body T tc tc2 P A err H
  for T :: "('f::show,'t::show,'d) theory"
  and tc :: "('h::show,'f,'v::show,'t) theory_checker"
  and tc2
  and P :: "('f,'v,'t,'l::show,'tr::show) lts"
  and A
  and err
  and H +
  assumes ans: "set (nodes H) = art.nodes A"
  and trans_P_list: "\<And> l. set (trans_P_list H l) =
    {(\<tau>,trans_target \<tau>') | \<tau> \<tau>'. \<tau> \<in> transitions P \<and> \<tau>' = transition_of P \<tau>  \<and> trans_source \<tau>' = l}"
begin

lemma initial_cond_checker[iff]: "isOK(initial_cond_checker) = initial_cond"
  unfolding initial_cond_def initial_cond_checker_def by auto

lemma error_cond_checker[iff]: "isOK(error_cond_checker) \<Longrightarrow> error_cond"
  unfolding error_cond_def error_cond_checker_def
  by (auto simp: ans zip_Cons split: option.splits dest: tc.formula_checker)

lemma cover_edges_cond_checker[iff]: "isOK(cover_edges_cond_checker) \<Longrightarrow> cover_edges_cond"
  unfolding cover_edges_cond_def cover_edges_cond_checker_def
  by (auto simp: ans split: art_edge.splits option.splits dest: tc.formula_checker)

lemma simulation_cond_checker[intro]: assumes "isOK(simulation_cond_checker)"
  shows simulation_cond
proof -
  note ok = assms[unfolded simulation_cond_checker_def, simplified, unfolded Let_def ans]
  show ?thesis
    unfolding simulation_cond_def
  proof (intro conjI allI impI)
    fix a tas l \<tau> \<phi> l'
    assume *: "a \<in> art.nodes A" "edge A a = Children tas" "l = node_location A a"
       "\<tau> \<in> transitions P" "transition_of P \<tau> = Transition l \<phi> l'"
    from conjunct2[OF ok, rule_format, OF *(1)] *(2-)
    show "\<exists>b\<in>art.nodes A. (\<tau>, b) \<in> set tas \<and> node_location A b = l'" by (fastforce simp: trans_P_list)
  qed (insert ok, auto)
qed

lemma children_edges_cond_checker[intro]: assumes "isOK(children_edges_cond_checker)"
  shows "children_edges_cond"
  unfolding children_edges_cond_def 
proof (intro allI impI)
  fix a tas \<tau> b
  assume a: "a \<in> art.nodes A" 
    and edge: "edge A a = Children tas" and mem: "(\<tau>, b) \<in> set tas"
  note ok = assms[unfolded children_edges_cond_checker_def, simplified, unfolded ans, rule_format, OF a, 
    unfolded edge, simplified]
  from ok obtain hints where tr: "transition_hints H a = Some hints" and len: "length hints = length tas"
    by (auto split: option.splits)
  from mem len obtain h where mem: "((\<tau>,b),h) \<in> set (zip tas hints)" unfolding set_zip
    by (auto simp: set_conv_nth)
  with ok[unfolded tr, simplified]
  show "transition_entailment P (node_invariant A a) \<tau> (node_invariant A b)"
    unfolding transition_entailment_def
    by (auto intro!: tc2.formula_checker)
qed

lemma art_safety_checker_main:
  "safety_prog (Safety_Prog err P) \<Longrightarrow> isOK(art_safety_checker_main) \<Longrightarrow> safe (Safety_Prog err P)"
  by (intro well_formed_sound, auto simp: art_safety_checker_main_def well_formed_def)

end

datatype ('tr,'h) art_edge_impl =
  Cover_Edge art_node 'h
| Children_Edge "('tr \<times> art_node \<times> 'h) list"

datatype ('f,'v,'t,'l,'tr,'h) art_node_impl = Art_Node
  (name: art_node) (invariant: "('f,'v,'t) formula") (location: 'l) (edge: "('tr,'h) art_edge_impl")

fun is_cover_node :: "('f,'v,'t,'l,'tr,'h) art_node_impl \<Rightarrow> bool" where
  "is_cover_node (Art_Node _ _ _ (Cover_Edge _ _)) = True"
| "is_cover_node (Art_Node _ _ _ (Children_Edge _)) = False"

record ('f,'v,'t,'l,'tr,'h) art_impl =
  initial_node :: art_node
  nodes :: "('f,'v,'t,'l,'tr,'h) art_node_impl list"
  error_hints :: "(art_node \<times> 'h) list"

definition art_nodes :: "('f,'v,'t,'l,'tr,'h) art_impl \<Rightarrow> art_node list" where
  "art_nodes Ai = map name (nodes Ai)"

fun art_edge_of :: "('tr,'h) art_edge_impl \<Rightarrow> 'tr art_edge" where
  "art_edge_of (Cover_Edge an _) = Cover an"
| "art_edge_of (Children_Edge ans) = Children (map (\<lambda> (t,a,_). (t,a)) ans)"

definition art_of :: "('f,'v,'t,'l,'tr,'h) art_impl \<Rightarrow> ('f,'v,'t,'l,'tr) art" where
  "art_of Ai \<equiv>
    let i = art_impl.initial_node Ai; ans = art_impl.nodes Ai in \<lparr>
      art.initial_node = i,
      art.nodes = set (art_nodes Ai),
      art.edge = (the o map_of (map (\<lambda> a. (name a, art_edge_of (edge a))) ans)),
      node_location = (the o map_of (map (\<lambda> a. (name a, location a)) ans)),
      node_invariant = (the o map_of (map (\<lambda> a. (name a, invariant a)) ans))
    \<rparr>"

lemma nodes_art_of[simp]: "art.nodes (art_of Ai) = set (art_nodes Ai)" by (simp add: art_of_def Let_def)

definition error_hints :: "('f,'v,'t,'l,'tr,'h) art_impl \<Rightarrow> art_node \<Rightarrow> 'h option" where
  "error_hints Ai = map_of (art_impl.error_hints Ai)"

definition cover_hints :: "('f,'v,'t,'l,'tr,'h) art_impl \<Rightarrow> art_node \<Rightarrow> 'h option" where
  "cover_hints Ai =
    map_of (map (\<lambda> a. (name a, case edge a of Cover_Edge _ h \<Rightarrow> h))
      (filter is_cover_node (art_impl.nodes Ai)))"

definition transition_hints :: "('f,'v,'t,'l,'tr,'h) art_impl \<Rightarrow> art_node \<Rightarrow> 'h list option" where
  "transition_hints Ai =
    map_of (map (\<lambda> a. (name a, case edge a of Children_Edge cs \<Rightarrow> map (\<lambda> (_,_,h). h) cs))
      (filter (\<lambda> a. \<not> is_cover_node a) (art_impl.nodes Ai)))"

context pre_art_checker begin

definition "make_hinter Pi Ai P \<equiv> \<lparr>
  hinter.nodes = art_nodes Ai,
  trans_P_list = succ_transitions Pi P,
  error_hints = error_hints Ai,
  cover_hints = cover_hints Ai,
  transition_hints = transition_hints Ai
\<rparr>"

definition art_safety_checker ::
  "('f, 'v, 't, 'l::show, 'tr::show) lts_impl \<Rightarrow> ('f,'v,'t,'l,'tr,'h list) art_impl \<Rightarrow> 'l \<Rightarrow> shows check" where
  "art_safety_checker Pi Ai err \<equiv>
   let P = lts_of Pi;
       A = art_of Ai;
       H = make_hinter Pi Ai P
   in
   pre_art_checker_body.art_safety_checker_main T tc tc2 P A err H <+?
   (\<lambda> s. shows ''could not ensure safety w.r.t. error location '' o shows err o shows '' via ART:\<newline>'' o s)"

end

context art_checker begin

lemma art_safety_checker:
  assumes Pi: "lts_impl Pi" and ok: "isOK (art_safety_checker Pi Ai err)"
  shows "safe (Safety_Prog err (lts_of Pi))" (is "safe (Safety_Prog _ ?P)")
proof -
  interpret art_checker_body T tc tc2 "?P" "art_of Ai" err "make_hinter Pi Ai ?P"
    by (standard, auto simp: make_hinter_def succ_transitions)
  show ?thesis
    apply(rule art_safety_checker_main)
    using Pi ok[unfolded art_safety_checker_def] by auto
qed
end

term pre_art_checker.art_safety_checker
thm art_checker.art_safety_checker

end
