(*
Author:  Alexander Krauss <krauss@in.tum.de> (2009-2011)
Author:  Christian Sternagel <c.sternagel@gmail.com> (2010-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
License: LGPL (see file COPYING.LESSER)
*)

signature ISAFOR_TACTICS =
sig
  val add_constr_sims : string -> theory -> theory
  val reduce_tac : Proof.context -> term -> int -> tactic
  val reduce_dp_tac : Proof.context -> term -> int -> tactic

  val derive_sim_lemma_any : Proof.context -> thm list -> thm
end

structure Isafor_Tactics(*: ISAFOR_TACTICS*) =
struct

open Isafor_Basics
open Call_Tree

val supt_const = Const (@{const_name Term.supt}, trsT)

val tracing = false
fun trace_tac ctxt msg = if tracing then print_tac ctxt msg else all_tac
fun trace_tac' ctxt = K o (trace_tac ctxt)

fun rtrace_tac ctxt rl i =
  rtac rl i
  ORELSE (fn st =>
    let
      val thy = Thm.theory_of_thm st
      val _ = Output.tracing (Syntax.string_of_term_global thy (Thm.prop_of rl))
    in
      (print_tac ctxt "FAILED" THEN (fn _ => raise Fail "stopped")) st
    end)

(*rewrites both sides of an equality (but = stays intact), or the
whole term if it is not an equality.*)
fun lhs_rhs_conv cv ct =
  case Thm.term_of ct of
    Const (@{const_name HOL.Trueprop}, _) $ (Const (@{const_name HOL.eq}, _) $ _ $ _) =>
      Conv.arg_conv (Conv.binop_conv cv) ct
  | _ => cv ct

(*apply constant substitutions only at the correct positions, keeping structure intact*)
fun subst_conv ctxt =
  let
    val subst_eqs = map mk_meta_eq (Trs_Data.Subst.get ctxt)
    val cv = Raw_Simplifier.rewrite ctxt false subst_eqs
  in
    Conv.params_conv ~1 (K
      ((Conv.concl_conv ~1 cv)
       then_conv (Conv.prems_conv ~1 (lhs_rhs_conv cv)))) ctxt
  end

fun emb_simp_tac ctxt extra =
  simp_tac (put_simpset HOL_basic_ss ctxt addsimps (extra @ Trs_Data.get_embs ctxt @ 
     @{thms TermFun_Setup.map_constructor_Fun map_simps
     }))

fun sub_rewrite_list_tac tacs i =
  args_qrsteps_tac (fn j => nth tacs (j - i) j) i

fun sub1_rewrite_tac tac i =
  args_qrsteps_tac (fn j => if i = j
    then tac j
    else rtac @{thm Transitive_Closure.rtrancl.rtrancl_refl} j) i

fun strip_cases ctac = ctac #> Seq.map snd

fun case_split_tac ctxt sub_tac = SUBGOAL (fn (t, i) =>
  let
    val case_arg =
      (dest_embed o dest_mapc o hd o snd o dest_app o fst o fst o dest_inrel) t
  in
    ((DETERM o strip_cases o Induct.cases_tac ctxt false [[SOME case_arg]] NONE [])
     THEN_ALL_NEW sub_tac) i
  end)

fun apply_sim_rule_tac ctxt = SUBGOAL (fn (t, i) =>
  let
    val ((l,_) , rel) = dest_inrel t
    val _ $ (_ $ Const (R, _)) = (Envir.beta_eta_contract rel)
    val (fname, _) = dest_app l
    val _ = Output.tracing ("name is " ^ fname)
  in
    rtac (Trs_Data.sim_for_trs ctxt fname R) i
  end)

fun check_constructor_tac ctxt = SOLVED' (simp_tac (put_simpset termfun_ss ctxt addsimps (Trs_Data.get_embs ctxt)))

fun simulation_tac ctxt _ _ (Variable _) = 
      (fn i => trace_tac ctxt ("sim_Var "^string_of_int i))
      THEN' rtac @{thm Transitive_Closure.rtrancl.rtrancl_refl}
  | simulation_tac ctxt _ _ (Unsupported (_, x, _)) =
      (fn i => trace_tac ctxt ("sim_Unsup- "^x^" "^string_of_int i))
      THEN' rtac @{thm Transitive_Closure.rtrancl.rtrancl_refl}
  | simulation_tac ctxt hyps ihyps (App (_, _, _, _, ts')) =
      (fn i => trace_tac ctxt ("sim_App "^string_of_int i))
      THEN' rtac @{thm Transitive_Closure.rtrancl_trans}
      THEN' (fn i => trace_tac ctxt ("sim_Appa "^string_of_int i))
      THEN' sub_rewrite_list_tac (map (simulation_tac ctxt hyps ihyps) ts')
      THEN' (fn i => trace_tac ctxt ("sim_Appb "^string_of_int i))
      THEN' ((resolve_tac ctxt ihyps THEN_ALL_NEW (trace_tac' ctxt ("sim_Appb1") THEN' 
        Simplifier.simp_tac (put_simpset HOL_basic_ss ctxt addsimps @{thms HOL.eq_True HOL.eq_False}) (* call simplify 
         to handle different induction rules of case bool and if bool *) 
        THEN' resolve_tac ctxt hyps))(*FIXME: modulo subst*)
        ORELSE' apply_sim_rule_tac ctxt)
      THEN' (fn i => trace_tac ctxt ("sim_Appc "^string_of_int i))
  | simulation_tac ctxt hyps ihyps (Case (_, dt, fname, _, arg', ts)) =
      let
        val trs_rules = Trs_Data.get_rules (Proof_Context.theory_of ctxt) (def_prefix ^ fname)
        val case_rewrites = #case_rewrites dt
        fun subcase_tac i j =
          let
            fun select_from ts = nth ts (j - i)
          in
            Subgoal.FOCUS (fn {context=ctxt,prems=hyps',...} =>
              trace_tac ctxt ("sim_Case: " ^ PolyML.makestring hyps' ^ "  and  " ^ PolyML.makestring hyps)
              THEN emb_simp_tac ctxt hyps' 1
              THEN trace_tac ctxt "sim_Case1"
              THEN rtac @{thm Transitive_Closure.converse_rtrancl_into_rtrancl} 1
              THEN trace_tac ctxt "sim_Case2"
              THEN (rtac (select_from trs_rules) THEN_ALL_NEW check_constructor_tac ctxt) 1
              THEN trace_tac ctxt "sim_Case3"
              THEN rewrite_goal_tac ctxt [mk_meta_eq (select_from case_rewrites)] 1
              THEN trace_tac ctxt "sim_Case4"
              THEN simulation_tac ctxt (hyps @ hyps') ihyps (#3 (select_from ts)) 1) ctxt j
              THEN trace_tac ctxt "sim_Case5"
          end
      in
        rtac @{thm Transitive_Closure.rtrancl_trans}
        THEN' sub1_rewrite_tac (simulation_tac ctxt hyps ihyps arg')
        THEN' (fn i => case_split_tac ctxt (subcase_tac i) i)
      end
  | simulation_tac ctxt hyps ihyps (If (_, fname, _, cond, thn, els)) =
      let
        val trs_rules = Trs_Data.get_rules (Proof_Context.theory_of ctxt) (def_prefix ^ fname)
        val case_rewrites = [@{thm HOL.if_True}, @{thm HOL.if_False}]
        val subtrees = [thn, els]
        fun subcase_tac i j =
          let
            fun select_from ts = nth ts (j - i)
          in
            Subgoal.FOCUS (fn {context=ctxt,prems=hyps',...} =>
              emb_simp_tac ctxt hyps' 1
              THEN rtac @{thm Transitive_Closure.converse_rtrancl_into_rtrancl} 1
              THEN trace_tac ctxt "sim_If 0"
              THEN (rtac (select_from trs_rules) THEN_ALL_NEW check_constructor_tac ctxt) 1
              THEN trace_tac ctxt "sim_If 1" 
              THEN rewrite_goal_tac ctxt [mk_meta_eq (select_from case_rewrites)] 1
              THEN trace_tac ctxt "sim_If 2"
              THEN simulation_tac ctxt (hyps @ hyps') ihyps (select_from subtrees) 1) ctxt j
          end
      in
        trace_tac' ctxt "sim_If A"
        THEN' rtac @{thm Transitive_Closure.rtrancl_trans}
        THEN' trace_tac' ctxt "sim_If B"
        THEN' sub1_rewrite_tac (simulation_tac ctxt hyps ihyps cond)
        THEN' trace_tac' ctxt "sim_If C"
        THEN' (fn i => case_split_tac ctxt (subcase_tac i) i)
        THEN' trace_tac' ctxt "sim_If D"
      end



fun simulation_prop con ctxt (Const (n, fT)) arity =
  let
    val thy = Proof_Context.theory_of ctxt
    val trs = case Trs_Data.get_def thy n of
      SOME (c, _) => Const (c, trsT)
    | NONE => Free ("R", trsT) (* FIXME *)
    val fT' = restrict_sortsT fT
    val (argTs, _) = chop arity (binder_types fT')
    val vars = Variable.variant_frees ctxt [] (map (pair "x") argTs)
      |> map Free
    val lhs = mk_app con true n (map ((mk_mapc oo maybe_mk_embed) thy) vars)
    val rhs = (mk_mapc oo maybe_mk_embed) thy (list_comb (Const (n, fT'), vars))
    val prop = mk_inrel (lhs, rhs) (mk_star (mk_qrstep mk_default_q trs))
  in
    (vars, prop)
  end

local open Conv
in

(* apply rules on both sides, then simulation_tac *)
fun unfold_and_sim_tac ctxt ihyps ((trs_rule, simp_rule), tree) =
  let
    val ihyps' = map (Conv.fconv_rule (subst_conv ctxt)) ihyps
    val _ = Output.tracing (@{make_string} trs_rule)
    val _ = Output.tracing (@{make_string} simp_rule)
    val _ = Output.tracing (@{make_string} ihyps)
    val _ = Output.tracing (@{make_string} ihyps')
  in
    rtac @{thm Transitive_Closure.converse_rtrancl_into_rtrancl}
    THEN' trace_tac' ctxt "before emb_simp"
    THEN' emb_simp_tac ctxt []
    THEN' trace_tac' ctxt "before rule"
    THEN' (rtac trs_rule THEN_ALL_NEW (trace_tac' ctxt "mid rule" THEN' check_constructor_tac ctxt))
    THEN' trace_tac' ctxt "after rule"
    THEN' CONVERSION ((arg_conv o arg1_conv o arg_conv o arg_conv o arg_conv o rewr_conv)
      (mk_meta_eq simp_rule))
    THEN' trace_tac' ctxt "after conv"
    THEN' simulation_tac ctxt [] ihyps' tree
  end
end

fun derive_sim_lemma ctxt f arity tac =
  let
    val (f', ctxt') = yield_singleton Variable.importT_terms f ctxt
    val (vars, goal) = simulation_prop false ctxt' f' arity
  in
    Goal.prove ctxt' (map (fst o dest_Free) vars) [] goal
      (fn {context=ctxt, ...} => tac ctxt vars)
    |> singleton (Proof_Context.export ctxt' ctxt)
  end

fun derive_sim_lemma_fun ctxt (f as Const (n, _)) arity ind_rule simps =
  let
    val thy = Proof_Context.theory_of ctxt
    val rules = (Trs_Data.get_rules thy (def_prefix ^ n) ~~ simps)
      ~~ (Trs_Data.get_trees thy n |> map #2)

    fun tac ctxt vars =
      ((DETERM o strip_cases o Induct.induct_tac ctxt false
         [map (fn t => SOME (NONE, (t, false))) vars] [] [] (SOME [ind_rule]) [])
       THEN_ALL_NEW (fn i => Subgoal.FOCUS (fn {context=ctxt, prems=ihyps, ...} =>
         unfold_and_sim_tac ctxt ihyps (nth rules (i - 1)) 1) ctxt i)) 1
  in
    derive_sim_lemma ctxt f arity tac
  end

fun derive_sim_lemma_def ctxt (f as Const (n, _)) arity def =
  let  
    val thy = Proof_Context.theory_of ctxt
    val rules = ((the_single (Trs_Data.get_rules thy (def_prefix ^ n)), def),
      #2 (the_single (Trs_Data.get_trees thy n)))
  in
    derive_sim_lemma ctxt f arity (fn ctxt => fn _ =>
      unfold_and_sim_tac ctxt [] rules 1)
  end

fun derive_sim_lemma_primrec ctxt (f as Const (n, _)) arity simps =
  let
    val thy = Proof_Context.theory_of ctxt
    val rules = (Trs_Data.get_rules thy (def_prefix ^ n) ~~ simps)
      ~~ (Trs_Data.get_trees thy n |> map #2)
    val _ $ (_ $ lhs $ _) = Thm.prop_of (hd simps)
    val (_, args) = strip_comb lhs
    val induction_pos = find_index (not o can dest_Var) args
    val Type (tname, _) = fastype_of (nth args induction_pos)
    val ind_rule = #induct (Old_Datatype_Data.the_info thy tname)
    fun other_vars vs =
      Library.take induction_pos vs
      @ Library.drop (induction_pos + 1) vs

    fun tac ctxt vars =        
      (((DETERM o strip_cases o Induct.induct_tac ctxt false
         [[SOME (NONE, (nth vars induction_pos, false))]]
         [map dest_Free (other_vars vars)] [] (SOME [ind_rule]) [])
       THEN' K (print_tac ctxt "after induct"))
       THEN_ALL_NEW (fn i => Subgoal.FOCUS (fn {context=ctxt, prems=ihyps, ...} =>
         unfold_and_sim_tac ctxt ihyps (nth rules (i - 1)) 1) ctxt i)) 1
  in
    derive_sim_lemma ctxt f arity tac
  end

datatype spec_format =
  Def
  | Primrec of int * Old_Datatype.info
  | Fun of thm

(* Is it fun, definition or primrec? Crude heuristic!!! *)
fun guess_spec_format ctxt simps =
  let
    val thy = Proof_Context.theory_of ctxt
    (* FIXME:
      - sometimes exception "Bind" is raised without (no_sim) for "define_trs"
      - find reason and provide informative error message
    *)
    val _ $ (_ $ lhs $ _) = Thm.prop_of (hd simps)
    val (f, args) = strip_comb lhs
    val arity = length args
    val spec_format =
      case Function_Common.import_function_data f ctxt of
        SOME ({defname,...}) => 
          Fun (Global_Theory.get_thm thy (defname ^ ".induct"))
      | NONE => case find_index (not o can dest_Var) args of
         ~1 => Def
       | i => let val Type (tname, _) = fastype_of (nth args i)
              in
                Primrec (i, Old_Datatype_Data.the_info thy tname)
              end
  in
    (f, arity, spec_format)
  end

fun derive_sim_lemma_any ctxt simps =
  let
    val (f, arity, spec_format) = guess_spec_format ctxt simps
  in
    case spec_format of
      Def => derive_sim_lemma_def ctxt f arity (the_single simps)
    | Fun ind_rule => derive_sim_lemma_fun ctxt f arity ind_rule simps
    | Primrec (_, _) => derive_sim_lemma_primrec ctxt f arity simps
  end

fun add_constr_sims tname thy =
  let
    val ctxt = Proof_Context.init_global thy
    val constrs = the (Old_Datatype_Data.get_constrs thy tname)

    fun mk_sim (c, T) =
      let
        val arity = length (binder_types T)
        val trs = Free ("R", trsT) (* FIXME *)
        val (vars, goal) = simulation_prop true ctxt (Const (c, T)) arity
      in
        Goal.prove ctxt (map (fst o dest_Free) (trs :: vars)) [] goal
          (fn {prems=_, context=ctxt} =>
            emb_simp_tac ctxt [] 1
              THEN rtac @{thm Transitive_Closure.rtrancl.rtrancl_refl} 1)
        |> Thm.close_derivation
      end
    val sim_rules = map mk_sim constrs
  in
    Context.theory_map (fold Trs_Data.add_sim sim_rules) thy
  end

fun mk_termination_rel trs fname argTs =
  let
    fun apply cn t = Const (cn, trsT --> trsT) $ t

    val R =
      trs
      |> mk_qrstep mk_default_q
      |> (fn t => HOLogic.mk_binop union_const_name (supt_const, t))
      |> apply @{const_name Transitive_Closure.trancl}
      |> apply @{const_name Relation.converse}

    val vars = map_index (fn (i,T) => Var (("x", i), T)) argTs
    
    val func = HOLogic.tupled_lambda 
      (foldr1 HOLogic.mk_prod vars) (mk_app false true fname (map (mk_mapc o mk_embed) vars))
    val argsT = foldr1 HOLogic.mk_prodT argTs
  in
    Const (@{const_name Relation.inv_image}, trsT --> (argsT --> termT)
      --> HOLogic.mk_setT (HOLogic.mk_prodT (argsT, argsT))) $ R $ func
  end

fun mk_dp_termination_rel dps trs fname argTs =
  let
    fun apply cn t = Const (cn, trsT --> trsT) $ t

    val R =
      mk_dprstep dps trs
      |> apply @{const_name Transitive_Closure.trancl}
      |> apply @{const_name Relation.converse}

    val vars = map_index (fn (i,T) => Var (("x", i), T)) argTs
    
    val func = HOLogic.tupled_lambda 
      (foldr1 HOLogic.mk_prod vars) (mk_sharp fname (map (mk_mapc o mk_embed) vars))
    val argsT = foldr1 HOLogic.mk_prodT argTs
  in
    Const (@{const_name Relation.inv_image}, trsT --> (argsT --> termT)
      --> HOLogic.mk_setT (HOLogic.mk_prodT (argsT, argsT))) $ R $ func
  end

fun rewrite_step_tac ctxt rrule i : tactic =
  trace_tac ctxt "rewrite_step:"
  THEN rtac @{thm TermFun_Setup.qterm_relI_step} i
  THEN emb_simp_tac ctxt [] i
  THEN trace_tac ctxt "rewstepA"
  THEN (rtrace_tac ctxt rrule THEN_ALL_NEW check_constructor_tac ctxt) i
  THEN trace_tac ctxt "rewstepB"

fun finish_tac ctxt trees = 
  (fn i => trace_tac ctxt ("finish1 "^string_of_int i))
  THEN' sub_rewrite_list_tac (map (simulation_tac ctxt [] []) trees)
  THEN' (fn i => trace_tac ctxt ("finish2 "^string_of_int i))
  THEN' rtac @{thm Transitive_Closure.rtrancl.rtrancl_refl}

fun finish_tac_trs ctxt trees =
  rtac @{thm TermFun_Setup.qterm_relI_steps}
  THEN' finish_tac ctxt trees

fun finish_tac_dp ctxt trees =
  rtac @{thm TermFun_Setup.dp_qterm_relI_steps}
  THEN' finish_tac ctxt trees

fun list_member_tac 0 = rtac @{thm TermFun_Setup.list_memberI1}
  | list_member_tac m = rtac @{thm TermFun_Setup.list_memberI2}
                          THEN' list_member_tac (m - 1)

fun drop_head_tac (j : int) : int -> tactic=
  rtac @{thm TermFun_Setup.qterm_relI_sub}
  THEN' rtac @{thm Term.supt.arg}
  THEN' list_member_tac j

(* tactics that consume extra hypotheses *)
type htactic = thm list -> int -> tactic

val KEEP_HYPS : (int -> tactic) -> htactic -> htactic =
  fn tac => fn htac => fn hyps =>
    tac THEN' htac hyps

val CONSUME_HYP : (thm -> int -> tactic) -> htactic -> htactic =
  fn tac => fn htac => fn hyp :: hyps =>
    tac hyp THEN' htac hyps

fun NO_HYPS tac [] = tac
  | NO_HYPS _ _ = error "too many hyps"

fun unfold_case_tac ctxt rule arg hyp =
  trace_tac' ctxt (cat_lines ["at case with rule/hyp ", 
    Syntax.string_of_term ctxt (Thm.prop_of rule),
    Syntax.string_of_term ctxt (Thm.prop_of hyp)])
  THEN' rtac @{thm TermFun_Setup.qterm_relI_steps}
  THEN' sub1_rewrite_tac (simulation_tac ctxt [] [] arg)
  THEN' trace_tac' ctxt "caseB"
  THEN' simp_tac (put_simpset HOL_basic_ss ctxt addsimps [hyp])
  THEN' trace_tac' ctxt "caseC"
  THEN' rewrite_step_tac ctxt rule
  THEN' trace_tac' ctxt "caseD"

fun reduce_tac ctxt (Const (fname, _)) = SUBGOAL (fn (t, i) =>
  let
    val thy = Proof_Context.theory_of ctxt
    val _ $ (_ $ Abs(_, argT, _)) = t
    val typs = HOLogic.strip_tupleT argT (* FIXME *)
    val trs = Const (fst (the (Trs_Data.get_def thy fname)), trsT)
    val main_rules = Trs_Data.get_rules thy (def_prefix ^ fname)
    val trees = Trs_Data.get_trees thy fname |> map #2
    val get_rule = nth o Trs_Data.get_rules thy o (fn f => def_prefix ^ f)

    fun rec_tacs (Variable _) : htactic list = []
      | rec_tacs (Unsupported _) : htactic list = []
      | rec_tacs (App (_, typ, _, _, ts)) =
          forall_subtrees (fn i => map (KEEP_HYPS (drop_head_tac i))) ts
          |> (if typ = Rec then cons (NO_HYPS (finish_tac_trs ctxt ts)) else I)
      | rec_tacs (Case (_, _, f, _, arg, cs)) =
          map (KEEP_HYPS (drop_head_tac 0)) (rec_tacs arg)
          @ forall_subtrees (fn i => map (CONSUME_HYP (unfold_case_tac ctxt (get_rule f i) arg))) (map #3 cs)
      | rec_tacs (If (_, f, _, cond, thn, els)) =
          map (KEEP_HYPS (drop_head_tac 0)) (rec_tacs cond)
          @ forall_subtrees (fn i => map (CONSUME_HYP (unfold_case_tac ctxt (get_rule f i) cond))) [thn,els]
    and forall_subtrees f =
      flat o map_index (uncurry f o apsnd rec_tacs)

    val rules_tacs =
      maps (fn (rule, tree) => map (pair rule) (rec_tacs tree)) (main_rules ~~ trees)

    fun simplify_subgoal_tac 1 =
      rtac @{thm Wellfounded.wf_inv_image} 1
      THEN rtac @{thm Abstract_Rewriting.SN_imp_wf} 1
      THEN rtac @{thm TermFun_Setup.SN_supt_qrstep_trancl} 1
    | simplify_subgoal_tac i =
      let
        val (rule, tac) = nth rules_tacs (i - 2)
      in
        trace_tac ctxt "initial goal" THEN
        CONVERSION (subst_conv ctxt) i
        THEN Subgoal.FOCUS (fn {context=ctxt,prems=hyps,...} =>
          trace_tac ctxt "A"
          THEN rtac @{thm TermFun_Setup.inv_imageI} 1
          THEN trace_tac ctxt "B"
          THEN rtac @{thm Relation.converseI} 1
          THEN trace_tac ctxt "C"
          THEN rewrite_goal_tac ctxt [mk_meta_eq @{thm Product_Type.split_conv}] 1
          THEN trace_tac ctxt "C1"
          THEN rtac @{thm TermFun_Setup.qterm_relI_initial} 1
          THEN trace_tac ctxt "D"
          THEN emb_simp_tac ctxt [] 1
          THEN trace_tac ctxt "E"
          THEN (rtac rule THEN_ALL_NEW check_constructor_tac ctxt) 1
          THEN trace_tac ctxt "F"
          THEN tac hyps 1) ctxt i
      end
in
  (Function_Relation.relation_tac ctxt (K (mk_termination_rel trs fname typs))
  THEN_ALL_NEW simplify_subgoal_tac) i
end)
  | reduce_tac _ _ = (K all_tac)

fun reduce_dp_tac ctxt (Const (fname, _)) = SUBGOAL (fn (t, i) =>
  let
    val thy = Proof_Context.theory_of ctxt
    val _ $ (_ $ Abs(_, argT, _)) = t
    val typs = HOLogic.strip_tupleT argT (* FIXME *)
    val (dpsN,_,trsN,_) = (the (Trs_Data.get_dp_def thy fname))
    val trs = Const (trsN,trsT)
    val dps = Const (dpsN,trsT)
    val main_dps = Trs_Data.get_dp_rules thy (sharp_prefix ^ fname)
    fun get_rule f i = 
       let val rules = Trs_Data.get_dp_rules thy (sharp_prefix ^ f)
           val rule = if i < length rules then nth rules i 
                      else error ("no " ^ string_of_int i^ "-th rule for " ^ f)
           in fst rule
       end
    fun rec_tacs (Variable _) : htactic list = []
      | rec_tacs (Unsupported _) : htactic list = []
      | rec_tacs (App (_, _, _, _, ts)) =
          [NO_HYPS (finish_tac_dp ctxt ts)]
      | rec_tacs (Case (_, _, f, _, arg, cs)) =
          (* map (KEEP_HYPS (drop_head_tac 0)) (rec_tacs arg) *)
          rec_tacs arg
          @ forall_subtrees (fn i => map (CONSUME_HYP (unfold_case_tac ctxt (get_rule f i) arg))) (map #3 cs)
      | rec_tacs (If (_, f, _, cond, thn, els)) =
          map (KEEP_HYPS (drop_head_tac 0)) (rec_tacs cond)
          @ forall_subtrees (fn i => map (CONSUME_HYP (unfold_case_tac ctxt (get_rule f i) cond))) [thn,els]
    and forall_subtrees f =
      flat o map_index (uncurry f o apsnd rec_tacs)

    val rules_tacs =
      maps (fn (rule, tree) => map (pair rule) (rec_tacs tree)) (main_dps)

    fun simplify_subgoal_tac 1 =
      rtac @{thm Wellfounded.wf_inv_image} 1
      THEN rtac @{thm Abstract_Rewriting.SN_imp_wf} 1
      THEN rtac @{thm TermFun_Setup.dp_SN_dpr_step_trancl} 1
      THEN rtac @{thm TermFun_Setup.finite_dpp_imp_SN_dpr_step} 1
    | simplify_subgoal_tac i =
      let
        val (rule, tac) = nth rules_tacs (i - 2)
      in
        trace_tac ctxt ("initial goal tried with " ^ @{make_string} rule) THEN
        CONVERSION (subst_conv ctxt) i
        THEN Subgoal.FOCUS (fn {context=ctxt,prems=hyps,...} =>
          trace_tac ctxt "A"
          THEN rtac @{thm TermFun_Setup.inv_imageI} 1
          THEN trace_tac ctxt "B"
          THEN rtac @{thm Relation.converseI} 1
          THEN trace_tac ctxt "C"
          THEN rewrite_goal_tac ctxt [mk_meta_eq @{thm Product_Type.split_conv}] 1
          THEN trace_tac ctxt "C1"
          THEN rtac @{thm TermFun_Setup.dp_qterm_relI_initial} 1
          THEN trace_tac ctxt "D"
          THEN emb_simp_tac ctxt [] 1
          THEN trace_tac ctxt "E"
          THEN (rtac rule THEN_ALL_NEW check_constructor_tac ctxt) 1
          THEN trace_tac ctxt "F"
          THEN tac hyps 1) ctxt i
      end
in
  (Function_Relation.relation_tac ctxt (K (mk_dp_termination_rel dps trs fname typs))
  THEN_ALL_NEW simplify_subgoal_tac) i
end)
  | reduce_dp_tac _ _ = (K all_tac)

end
